# Inspired by: https://github.com/mikeckennedy/jinja_partials
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from quart import Quart
    from starlette.templating import Jinja2Templates


def register_starlette(templates: "Jinja2Templates") -> None:
    def render_partial(name: str, **context: Any) -> str:
        return templates.get_template(name).render(**context)

    templates.env.globals.update(render_partial=render_partial)


def register_quart(app: "Quart") -> None:
    async def render_partial(name: str, **context: Any) -> str:
        # NOTE: Quart automatically awaits function calls in its templates environment.
        return await app.jinja_env.get_template(name).render_async(**context)

    app.jinja_env.globals.update(render_partial=render_partial)



import httpx
import pytest
from starlette.applications import Starlette
from starlette.middleware import Middleware
from starlette.responses import JSONResponse, Response
from starlette.routing import Route

from asgi_htmx import HtmxDetails, HtmxMiddleware
from asgi_htmx import HtmxRequest as Request


async def index(request: Request) -> Response:
    htmx = request.scope["htmx"]
    assert isinstance(htmx, HtmxDetails)
    data = {
        "__bool__": bool(htmx),
        "boosted": htmx.boosted,
        "current_url": htmx.current_url,
        "history_restore_request": htmx.history_restore_request,
        "prompt": htmx.prompt,
        "target": htmx.target,
        "trigger": htmx.trigger,
        "trigger_name": htmx.trigger_name,
        "triggering_event": htmx.triggering_event,
    }
    return JSONResponse(data)


app = Starlette(
    routes=[Route("/", index)],
    middleware=[Middleware(HtmxMiddleware)],
)


@pytest.mark.asyncio
async def test_default() -> None:
    async with httpx.AsyncClient(app=app) as client:
        response = await client.get("http://testserver/")
        data = response.json()
        assert data["__bool__"] is False
        assert data["boosted"] is False
        assert data["current_url"] is None
        assert data["history_restore_request"] is False
        assert data["prompt"] is None
        assert data["target"] is None
        assert data["trigger"] is None
        assert data["trigger_name"] is None
        assert data["triggering_event"] is None


@pytest.mark.asyncio
async def test_bool_set() -> None:
    async with httpx.AsyncClient(app=app) as client:
        response = await client.get(
            "http://testserver/", headers={"HX-Request": "true"}
        )
        data = response.json()
        assert data["__bool__"] is True


@pytest.mark.asyncio
async def test_boosted_set() -> None:
    async with httpx.AsyncClient(app=app) as client:
        response = await client.get(
            "http://testserver/", headers={"HX-Boosted": "true"}
        )
        data = response.json()
        assert data["boosted"] is True


@pytest.mark.asyncio
async def test_current_url_set() -> None:
    async with httpx.AsyncClient(app=app) as client:
        response = await client.get(
            "http://testserver/",
            headers={"HX-Current-URL": "https://example.com"},
        )
        data = response.json()
        assert data["current_url"] == "https://example.com"


@pytest.mark.asyncio
async def test_current_url_set_urlencoded() -> None:
    async with httpx.AsyncClient(app=app) as client:
        response = await client.get(
            "http://testserver/",
            headers={
                "HX-Current-URL": "https%3A%2F%2Fexample.com%2F%3F",
                "HX-Current-URL-URI-AutoEncoded": "true",
            },
        )
        data = response.json()
        assert data["current_url"] == "https://example.com/?"


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "value, result",
    [
        ("false", False),
        ("true", True),
    ],
)
async def test_history_restore_request(value: str, result: bool) -> None:
    async with httpx.AsyncClient(app=app) as client:
        response = await client.get(
            "http://testserver/",
            headers={"HX-History-Restore-Request": value},
        )
        data = response.json()
        assert data["history_restore_request"] is result


@pytest.mark.asyncio
async def test_prompt_set() -> None:
    async with httpx.AsyncClient(app=app) as client:
        response = await client.get(
            "http://testserver/", headers={"HX-Prompt": "Yes, please"}
        )
        data = response.json()
        assert data["prompt"] == "Yes, please"


@pytest.mark.asyncio
async def test_target_set() -> None:
    async with httpx.AsyncClient(app=app) as client:
        response = await client.get(
            "http://testserver/", headers={"HX-Target": "clickme"}
        )
        data = response.json()
        assert data["target"] == "clickme"


@pytest.mark.asyncio
async def test_trigger_set() -> None:
    async with httpx.AsyncClient(app=app) as client:
        response = await client.get(
            "http://testserver/", headers={"HX-Trigger": "click"}
        )
        data = response.json()
        assert data["trigger"] == "click"


@pytest.mark.asyncio
async def test_trigger_name_set() -> None:
    async with httpx.AsyncClient(app=app) as client:
        response = await client.get(
            "http://testserver/", headers={"HX-Trigger-Name": "email"}
        )
        data = response.json()
        assert data["trigger_name"] == "email"


@pytest.mark.asyncio
async def test_triggering_event_bad_json() -> None:
    async with httpx.AsyncClient(app=app) as client:
        response = await client.get(
            "http://testserver/", headers={"Triggering-Event": "{"}
        )
        data = response.json()
        assert data["triggering_event"] is None


@pytest.mark.asyncio
async def test_triggering_event() -> None:
    async with httpx.AsyncClient(app=app) as client:
        response = await client.get(
            "http://testserver/",
            headers={
                "Triggering-Event": "%7B%22target%22%3A%20null%7D",
                "Triggering-Event-URI-AutoEncoded": "true",
            },
        )
        data = response.json()
        assert data["triggering_event"] == {"target": None}



from typing import Any, Literal, Protocol, overload

from ._models import HtmxDetails
from ._types import Scope

try:
    from starlette.requests import Request
except ImportError:  # pragma: no cover
    pass  # Starlette not installed.
else:

    class _HtmxScopeProto(Protocol):
        @overload
        def __getitem__(self, key: Literal["htmx"]) -> HtmxDetails:
            ...  # pragma: no cover

        @overload
        def __getitem__(self, key: str) -> Any:
            ...  # pragma: no cover

    class HtmxScope(_HtmxScopeProto, Scope):
        pass

    class HtmxRequest(Request):
        scope: HtmxScope



# Ripped from `starlette.types`.
import typing

Scope = typing.MutableMapping[str, typing.Any]
Message = typing.MutableMapping[str, typing.Any]

Receive = typing.Callable[[], typing.Awaitable[Message]]
Send = typing.Callable[[Message], typing.Awaitable[None]]

ASGIApp = typing.Callable[[Scope, Receive, Send], typing.Awaitable[None]]



from ._models import HtmxDetails
from ._types import ASGIApp, Receive, Scope, Send


class HtmxMiddleware:
    def __init__(self, app: ASGIApp) -> None:
        self._app = app

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] in ("http", "websocket"):
            scope["htmx"] = HtmxDetails(scope)

        await self._app(scope, receive, send)



import httpx
import pytest

pytest.importorskip("fastapi")


@pytest.mark.asyncio
async def test_example_fastapi() -> None:
    from examples.fastapi import app

    async with httpx.AsyncClient(app=app) as client:
        response = await client.get("http://testserver")
        assert response.status_code == 200



import json
from typing import Any, Optional
from urllib.parse import unquote

from ._types import Scope


class HtmxDetails:
    def __init__(self, scope: Scope) -> None:
        self._scope = scope

    def _get_header(self, name: bytes) -> Optional[str]:
        return _get_header(self._scope, name)

    def __bool__(self) -> bool:
        return self._get_header(b"HX-Request") == "true"

    @property
    def boosted(self) -> bool:
        return self._get_header(b"HX-Boosted") == "true"

    @property
    def current_url(self) -> Optional[str]:
        return self._get_header(b"HX-Current-URL")

    @property
    def history_restore_request(self) -> bool:
        return self._get_header(b"HX-History-Restore-Request") == "true"

    @property
    def prompt(self) -> Optional[str]:
        return self._get_header(b"HX-Prompt")

    @property
    def target(self) -> Optional[str]:
        return self._get_header(b"HX-Target")

    @property
    def trigger(self) -> Optional[str]:
        return self._get_header(b"HX-Trigger")

    @property
    def trigger_name(self) -> Optional[str]:
        return self._get_header(b"HX-Trigger-Name")

    @property
    def triggering_event(self) -> Any:
        value = self._get_header(b"Triggering-Event")

        if value is None:
            return None

        try:
            return json.loads(value)
        except json.JSONDecodeError:
            return None


def _get_header(scope: Scope, key: bytes) -> Optional[str]:
    key = key.lower()
    value: Optional[str] = None
    should_unquote = False

    for k, v in scope["headers"]:
        if k.lower() == key:
            value = v.decode("latin-1")
        if k.lower() == b"%s-uri-autoencoded" % key and v == b"true":
            should_unquote = True

    if value is None:
        return None

    return unquote(value) if should_unquote else value



from ._middleware import HtmxMiddleware
from ._models import HtmxDetails

try:
    from ._requests import HtmxRequest
except ImportError:  # pragma: no cover
    pass  # Starlette not installed.

__version__ = "0.1.0"

__all__ = [
    "__version__",
    "HtmxMiddleware",
    "HtmxDetails",
    "HtmxRequest",
]



from starlette.applications import Starlette
from starlette.middleware import Middleware
from starlette.responses import Response
from starlette.routing import Mount, Route
from starlette.staticfiles import StaticFiles
from starlette.templating import Jinja2Templates

from asgi_htmx import HtmxMiddleware
from asgi_htmx import HtmxRequest as Request

from .common import HERE, make_table
from .lib import render_partial

static = StaticFiles(directory=HERE / "static")
templates = Jinja2Templates(directory=HERE / "templates")
render_partial.register_starlette(templates)


async def home(request: Request) -> Response:
    return templates.TemplateResponse("home.html", {"request": request})


async def result(request: Request) -> Response:
    assert (htmx := request.scope["htmx"])
    template = "partials/result.html"
    context = {"request": request, "table": make_table(htmx)}
    return templates.TemplateResponse(template, context)


app = Starlette(
    routes=[
        Mount("/static", static, name="static"),
        Route("/", home, name="home"),
        Route("/result", result, name="result"),
    ],
    middleware=[Middleware(HtmxMiddleware)],
)



from typing import cast

from quart import Quart, render_template, request

from asgi_htmx import HtmxDetails

from .common import make_table
from .lib import render_partial

app = Quart(__name__)

render_partial.register_quart(app)


@app.route("/")
async def home() -> str:
    return await render_template("home.html", framework="quart")


@app.route("/result")
async def result() -> str:
    assert (htmx := HtmxDetails(cast(dict, request.scope)))
    context = {"table": make_table(htmx)}
    return await render_template("partials/result.html", **context)



from pathlib import Path

from asgi_htmx import HtmxDetails

HERE = Path(__file__).parent


def make_table(htmx: HtmxDetails) -> list:
    return [
        {"name": name, "value": getattr(htmx, name)}
        for name in dir(htmx)
        if not name.startswith("_")
    ]



from fastapi import FastAPI, Response
from fastapi.staticfiles import StaticFiles
from fastapi.templating import Jinja2Templates

from asgi_htmx import HtmxMiddleware
from asgi_htmx import HtmxRequest as Request

from .common import HERE, make_table
from .lib import render_partial

app = FastAPI()
app.mount("/static", StaticFiles(directory=HERE / "static"), name="static")
app.add_middleware(HtmxMiddleware)

templates = Jinja2Templates(directory=HERE / "templates")
render_partial.register_starlette(templates)


@app.get("/", name="home")
async def home(request: Request) -> Response:
    return templates.TemplateResponse("home.html", {"request": request})


@app.get("/result", name="result")
async def result(request: Request) -> Response:
    assert (htmx := request.scope["htmx"])
    template = "partials/result.html"
    context = {"request": request, "table": make_table(htmx)}
    return templates.TemplateResponse(template, context)
